from .closure import Add, Constant, Lambda, Multiply, One, Unity, Zero  # noqa: F401
from .container import (  # noqa: F401
    Broadcast,
    BroadcastReduce,
    Conditional,
    Parallel,
    ParallelDispatch,
    Reduce,
    Residual,
    Sequential,
)
from .conv import Conv1d, Conv2d, Conv3d  # noqa: F401
from .convert import continual, forward_stepping  # noqa: F401
from .delay import Delay  # noqa: F401
from .linear import Linear  # noqa: F401
from .module import CoModule, PaddingMode, TensorPlaceholder, call_mode  # noqa: F401
from .pooling import (  # noqa: F401
    AdaptiveAvgPool2d,
    AdaptiveAvgPool3d,
    AdaptiveMaxPool2d,
    AdaptiveMaxPool3d,
    AvgPool1d,
    AvgPool2d,
    AvgPool3d,
    MaxPool1d,
    MaxPool2d,
    MaxPool3d,
)
from .ptflops import _register_ptflops  # noqa: F401
from .shape import Reshape  # noqa: F401
from .utils import flat_state_dict, load_state_dict, state_dict  # noqa: F401



""" Register modules with conversion system and 3rd-party libraries """

from functools import wraps
from types import FunctionType
from typing import Callable, Type

from torch import Tensor, nn

from .closure import Lambda
from .container import Sequential
from .conv import Conv1d, Conv2d, Conv3d
from .linear import Linear
from .logging import getLogger
from .module import CallMode, CoModule, call_mode
from .pooling import (
    AdaptiveAvgPool2d,
    AdaptiveAvgPool3d,
    AdaptiveMaxPool2d,
    AdaptiveMaxPool3d,
    AvgPool1d,
    AvgPool2d,
    AvgPool3d,
    MaxPool1d,
    MaxPool2d,
    MaxPool3d,
)

logger = getLogger(__name__)


def forward_stepping(module: nn.Module, dim: int = 2):
    """Enhances torch.nn.Module with `forward_step` and `forward_steps`
    by unsqueezing the temporal dimension.

    NB: The passed module must not have time-dependent operations!
    For instance, `module = nn.Conv3d(1, 1, kernel_size=(1,1,1))` is OK,
    but results for `module = nn.Conv3d(1, 1, kernel_size=(3,3,3))` would be invalid.

    Alternatively, one may attempt to automatically convert the module by using
    `co.continual(module)` instead.

    Args:
        module (nn.Module): the torch.nn.Module to enchange
        dim (int, optional): The dimension to unsqueeze during `forward_step`. Defaults to 2.
    """

    def forward_step(func: Callable[[Tensor], Tensor]):
        @wraps(func)
        def call(x: Tensor, update_state=True) -> Tensor:
            x = x.unsqueeze(dim)
            x = func(x)
            x = x.squeeze(dim)
            return x

        return call

    def forward_steps(func: Callable[[Tensor], Tensor]):
        @wraps(func)
        def call(x: Tensor, pad_end=False, update_state=True) -> Tensor:
            x = func(x)
            return x

        return call

    def dummy(*args, **kwargs):
        ...  # pragma: no cover

    @staticmethod
    def build_from(mod):  # pragma: no cover
        return module.__class__()

    # module.__class__.__bases__ = (*module.__class__.__bases__, CoModule)
    orig_forward = module.forward
    module.forward_steps = forward_steps(module.forward)
    module.forward_step = forward_step(module.forward)
    module.delay = 0
    module.receptive_field = 1
    module.stride = getattr(module, "stride", 1)
    module.padding = getattr(module, "padding", 0)
    module.build_from = build_from
    module.get_state = dummy
    module.set_state = dummy
    module.clean_state = dummy

    # Call mode
    module.call_mode = CallMode.FORWARD

    def forward_with_callmode(*args, **kwargs):
        _call_mode = (
            call_mode.cur
            if call_mode.prev
            else getattr(module, "call_mode", CallMode.FORWARD)
        )
        if _call_mode == CallMode.FORWARD:
            return orig_forward(*args, *kwargs)
        return CoModule._call_impl(module, *args, **kwargs)

    module.forward = forward_with_callmode

    return module


# A mapping from torch.nn modules to continual modules
MODULE_MAPPING = {}

NAIVE_MAPPING = {
    # >> Activations
    nn.Threshold,
    nn.ReLU,
    nn.RReLU,
    nn.Hardtanh,
    nn.ReLU6,
    nn.Sigmoid,
    nn.Hardsigmoid,
    nn.Tanh,
    nn.SiLU,
    nn.Hardswish,
    nn.ELU,
    nn.CELU,
    nn.SELU,
    nn.GLU,  # has dim
    nn.GELU,
    nn.Hardshrink,
    nn.LeakyReLU,
    nn.LogSigmoid,
    nn.Softplus,
    nn.Softshrink,
    nn.PReLU,
    nn.Softsign,
    nn.Tanhshrink,
    nn.Softmin,  # has dim
    nn.Softmax,  # has dim
    nn.Softmax2d,
    nn.LogSoftmax,
    # >> Norm modules
    nn.BatchNorm1d,
    nn.BatchNorm2d,
    nn.BatchNorm3d,
    # >> Dropout modules
    nn.Dropout,
    nn.Dropout2d,
    nn.Dropout3d,
    nn.AlphaDropout,
    nn.FeatureAlphaDropout,
}


class ModuleNotRegisteredError(Exception):
    ...


def register(TorchNnModule: Type[nn.Module], CoClass: Type[CoModule]):
    CoModule._validate_class(CoClass)
    assert callable(getattr(CoClass, "build_from", None)), (
        f"To register {CoClass.__name__}, it should implement a `build_from` method:"
        """
        class MyContinualModule:

            @staticmethod
            def build_from(module: MyTorchModule) -> "MyContinualModule":
                ...
        """
    )
    MODULE_MAPPING[TorchNnModule] = CoClass
    return CoClass


def continual(module: nn.Module) -> CoModule:
    if type(module) in NAIVE_MAPPING:
        return forward_stepping(module)

    assert type(module) in MODULE_MAPPING, (
        f"A registered conversion for {module} was not found. "
        "You can register a custom conversion as follows:"
        """
        import continual as co

        co.convert.register(MyTorchModule, MyContinualModule)
        """
    )
    return MODULE_MAPPING[type(module)].build_from(module)


# Register modules with our conversion system

# Conv
register(nn.Conv1d, Conv1d)
register(nn.Conv2d, Conv2d)
register(nn.Conv3d, Conv3d)

# Pooling
register(nn.AvgPool1d, AvgPool1d)
register(nn.MaxPool1d, MaxPool1d)
register(nn.AvgPool2d, AvgPool2d)
register(nn.MaxPool2d, MaxPool2d)
register(nn.AdaptiveAvgPool2d, AdaptiveAvgPool2d)
register(nn.AdaptiveMaxPool2d, AdaptiveMaxPool2d)
register(nn.AvgPool3d, AvgPool3d)
register(nn.MaxPool3d, MaxPool3d)
register(nn.AdaptiveAvgPool3d, AdaptiveAvgPool3d)
register(nn.AdaptiveMaxPool3d, AdaptiveMaxPool3d)

# Linear
register(nn.Linear, Linear)

# Container
register(nn.Sequential, Sequential)

# Closure
register(FunctionType, Lambda)


# Register modules in `ptflops`
try:
    from ptflops import flops_counter as fc

    # Conv
    fc.MODULES_MAPPING[Conv1d] = fc.conv_flops_counter_hook
    fc.MODULES_MAPPING[Conv2d] = fc.conv_flops_counter_hook
    fc.MODULES_MAPPING[Conv3d] = fc.conv_flops_counter_hook

    # Pooling
    fc.MODULES_MAPPING[AvgPool1d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[MaxPool1d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[AvgPool2d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[MaxPool2d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[AdaptiveAvgPool2d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[AdaptiveMaxPool2d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[AvgPool3d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[MaxPool3d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[AdaptiveAvgPool3d] = fc.pool_flops_counter_hook
    fc.MODULES_MAPPING[AdaptiveMaxPool3d] = fc.pool_flops_counter_hook

except ModuleNotFoundError:  # pragma: no cover
    pass
except Exception as e:  # pragma: no cover
    logger.warning(f"Failed to add flops_counter_hook: {e}")



import torch
from torch import Tensor, nn
from torch.nn.functional import linear

from .module import CoModule


class Linear(CoModule, nn.Linear):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        device=None,
        dtype=None,
        channel_dim=-1,
    ) -> None:
        r"""Applies a linear transformation to the incoming data: :math:`y = xA^T + b`

        This module supports :ref:`TensorFloat32<tf32_on_ampere>`.

        Args:
            in_features: size of each input sample
            out_features: size of each output sample
            bias: If set to ``False``, the layer will not learn an additive bias.
                Default: ``True``
            channel_dim: Channel dimension index. Default: -1.

        Shape:
            - Input: :math:`(N, *, C_{in})` where :math:`*` means any number of
            additional dimensions and :math:`C_{in} = \text{infₑₐₜᵤᵣₑₛ}` if `channel_dim = -1`.
            - Output: :math:`(N, *, C_{out})` where all but the last dimension
            are the same shape as the input and :math:`C_{out} = \text{outfₑₐₜᵤᵣₑₛ}` if `channel_dim = -1`.

        Attributes:
            weight: the learnable weights of the module of shape
                :math:`(\text{out\_features}, \text{in\_features})`. The values are
                initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})`, where
                :math:`k = \frac{1}{\text{in\_features}}`
            bias:   the learnable bias of the module of shape :math:`(\text{out\_features})`.
                    If :attr:`bias` is ``True``, the values are initialized from
                    :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
                    :math:`k = \frac{1}{\text{in\_features}}`

        Examples::

            >>> m = nn.Linear(20, 30)
            >>> input = torch.randn(128, 20)
            >>> output = m(input)
            >>> print(output.size())
            torch.Size([128, 30])
        """
        nn.Linear.__init__(self, in_features, out_features, bias, device, dtype)
        self.channel_dim = channel_dim

    def extra_repr(self):
        return nn.Linear.extra_repr(self) + f", channel_dim={self.channel_dim}"

    def forward(self, input: Tensor) -> Tensor:
        if self.channel_dim != -1:
            input = input.swapaxes(self.channel_dim, -1)

        output = linear(input, self.weight, self.bias)  # Assumes channel-last

        if self.channel_dim != -1:
            output = output.swapaxes(self.channel_dim, -1)

        return output

    def forward_step(self, input: Tensor, update_state=True) -> Tensor:
        return self.forward(input)

    def forward_steps(self, input: Tensor, pad_end=False, update_state=True) -> Tensor:
        return self.forward(input)

    @staticmethod
    def build_from(
        module: nn.Linear,
        channel_dim=-1,
        **kwargs,
    ) -> "Linear":
        comodule = Linear(
            **{
                **dict(
                    in_features=module.in_features,
                    out_features=module.out_features,
                    bias=module.bias is not None,
                    device=module.weight.device,
                    dtype=module.weight.dtype,
                    channel_dim=channel_dim,
                ),
                **kwargs,
            }
        )
        with torch.no_grad():
            comodule.weight.copy_(module.weight)
            if module.bias is not None:
                comodule.bias.copy_(module.bias)
        return comodule



""" PyTorch Lightning compatible logging """

import logging
from functools import wraps


def _process_rank():
    try:
        import pytorch_lightning as pl

        if pl.utilities._HOROVOD_AVAILABLE:  # pragma: no cover
            import horovod.torch as hvd

            hvd.init()
            return hvd.rank()
        else:
            return pl.utilities.rank_zero_only.rank

    except ModuleNotFoundError:  # pragma: no cover
        return 0


process_rank = _process_rank()


def if_rank_zero(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        global process_rank
        if process_rank == 0:
            fn(*args, **kwargs)

    return wrapped


def getLogger(name):
    logger = logging.getLogger(name)
    logger._log = if_rank_zero(logger._log)
    return logger



import functools
import itertools
from abc import ABC
from enum import Enum
from typing import Any, Optional, Sequence, Tuple, Union

import torch
import torch.utils.hooks as hooks
from torch import Tensor
from torch.nn.modules.module import (
    _global_backward_hooks,
    _global_forward_hooks,
    _global_forward_pre_hooks,
)

from continual.utils import num_from


class TensorPlaceholder:
    shape: Tuple[int]

    def __init__(self, shape: Tuple[int] = tuple()):
        self.shape = shape

    def size(self):
        return self.shape

    def __len__(self):
        return 0


class PaddingMode(Enum):
    REPLICATE = "replicate"
    ZEROS = "zeros"


class CallMode(Enum):
    FORWARD = "forward"
    FORWARD_STEPS = "forward_steps"
    FORWARD_STEP = "forward_step"


class _CallModeContext(object):
    """Context-manager state holder."""

    def __init__(self):
        self.cur = CallMode.FORWARD
        self.prev = None

    def __call__(self, value: CallMode) -> "_CallModeContext":
        self.prev = self.cur
        self.cur = CallMode(value)
        return self

    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        self.cur = self.prev
        self.prev = None


call_mode = _CallModeContext()
"""Context-manager that holds a call_mode

When the call_mode context is used, the __call__ function of continual modules with be set accordingly

Example:
    >>> forward_output = module(forward_input)
    >>>
    >>> with co.call_mode("forward_step"):
    >>>     forward_step_output = module(forward_step_input)
"""


# First element must be a Tensor
State = Union[
    Tuple[Tensor, int],
    Tuple[Tensor, int, int],
]


def _clone_first(state: State) -> State:
    if state is None:
        return None
    return (state[0].clone(), *state[1:])


class CoModule(ABC):
    """Base class for continual modules.
    Deriving from this class enforces that neccessary class methods are implemented
    """

    def __init_subclass__(cls) -> None:
        CoModule._validate_class(cls)

    @staticmethod
    def _validate_class(cls):
        for fn, description in [
            ("forward_step", "forward computation for a single temporal step"),
            (
                "forward_steps",
                "forward computation for multiple temporal step",
            ),
            (
                "forward",
                "a forward computation which is identical to a regular non-continual forward.",
            ),
            ("get_state", "a retrieval of the internal state."),
            ("set_state", "an update of the internal state."),
            ("clean_state", "an internal state clean-up."),
        ]:
            assert callable(
                getattr(cls, fn, None)
            ), f"{cls} should implement a `{fn}` function which performs {description} to satisfy the CoModule interface."

        for prop in {"delay", "receptive_field"}:
            assert type(getattr(cls, prop, None)) in {
                int,
                property,
            }, f"{cls} should implement a `{prop}` property to satisfy the CoModule interface."

        for prop in {"stride", "padding"}:
            assert type(getattr(cls, prop, None)) in {
                int,
                property,
                tuple,
            }, f"{cls} should implement a `{prop}` property to satisfy the CoModule interface."

    @staticmethod
    def is_valid(module):
        try:
            CoModule._validate_class(module)
        except AssertionError:
            return False
        return True

    def get_state(self) -> Optional[State]:
        """Get model state

        Returns:
            Optional[State]: A State tuple if the model has been initialised and otherwise None.
        """
        ...  # pragma: no cover

    def set_state(self, state: State):
        """Set model state

        Args:
            state (State): State tuple to set as new internal internal state
        """
        ...  # pragma: no cover

    def clean_state(self):
        """Clean model state"""
        ...  # pragma: no cover

    receptive_field: int = 1
    stride: int = 1
    padding: int = 0
    make_padding = torch.zeros_like

    @property
    def delay(self) -> int:
        return self.receptive_field - 1 - num_from(self.padding)

    def forward_step(
        self, input: Tensor, update_state=True
    ) -> Union[Tensor, TensorPlaceholder]:
        """Forward computation for a single step with state initialisation

        Args:
            input (Tensor): Layer input.
            update_state (bool): Whether internal state should be updated during this operation.

        Returns:
            Union[Tensor, TensorPlaceholder]: Step output. This will be a placeholder while the module initialises and every (stride - 1) : stride.
        """
        state = self.get_state()
        if not update_state and state:
            state = _clone_first(state)
        output, state = self._forward_step(input, state)
        if update_state:
            self.set_state(state)
        return output

    def forward_steps(self, input: Tensor, pad_end=False, update_state=True) -> Tensor:
        """Forward computation for multiple steps with state initialisation

        Args:
            input (Tensor): Layer input.
            pad_end (bool): Whether results for temporal padding at sequence end should be included.
            update_state (bool): Whether internal state should be updated during this operation.

        Returns:
            Tensor: Layer output
        """
        outs = []
        tmp_state = self.get_state()

        if not update_state and tmp_state:
            tmp_state = _clone_first(tmp_state)

        for t in range(input.shape[2]):
            o, tmp_state = self._forward_step(input[:, :, t], tmp_state)
            if isinstance(o, Tensor):
                outs.append(o)

        if update_state:
            self.set_state(tmp_state)

        if pad_end:
            # Don't save state for the end-padding
            tmp_state = _clone_first(self.get_state()) or tmp_state
            for t, i in enumerate(
                [self.make_padding(input[:, :, -1]) for _ in range(self.padding[0])]
            ):
                o, tmp_state = self._forward_step(i, tmp_state)
                if isinstance(o, Tensor):
                    outs.append(o)

        if len(outs) == 0:
            return torch.tensor([])  # pragma: no cover

        return torch.stack(outs, dim=2)

    def forward(self, input: Tensor) -> Any:
        """Forward computation for multiple steps without state initialisation.
        This function is identical to the non-continual module found `torch.nn`

        Args:
            input (Tensor): Layer input.
        """
        ...  # pragma: no cover

    def warm_up(self, step_shape: Sequence[int]):
        """Warms up the model state with a dummy input.
        While this avoid the initial `self.delay` steps of `TensorPlaceholder` values,
        the initial `self.delay` steps will still produce inexact results.

        Args:
            step_shape (Sequence[int]): input shape with which to warm the model up, including batch size.
        """
        step_shape = (*step_shape[:2], self.delay, *step_shape[2:])
        dummy = self.make_padding(torch.zeros(step_shape, dtype=torch.float))
        self.forward_steps(dummy)

    _call_mode = CallMode.FORWARD

    @property
    def call_mode(self) -> CallMode:
        return self._call_mode

    @call_mode.setter
    def call_mode(self, value):
        self._call_mode = CallMode(value)
        if hasattr(self, "__len__"):
            for m in self:
                if hasattr(m, "call_mode"):
                    m.call_mode = self._call_mode

    def _call_impl(self, *input, **kwargs):  # noqa: C901  # pragma: no cover
        """Modified version torch.nn.Module._call_impl

        Returns:
            [type]: [description]
        """
        _call_mode = call_mode.cur if call_mode.prev else self.call_mode
        forward_call = {
            (True, CallMode.FORWARD): self._slow_forward,
            (False, CallMode.FORWARD): self.forward,
            (False, CallMode.FORWARD_STEPS): self.forward_steps,
            (False, CallMode.FORWARD_STEP): self.forward_step,
        }[(bool(torch._C._get_tracing_state()), _call_mode)]

        # If we don't have any hooks, we want to skip the rest of the logic in
        # this function, and just call forward.
        if not (
            self._backward_hooks
            or self._forward_hooks
            or self._forward_pre_hooks
            or _global_backward_hooks
            or _global_forward_hooks
            or _global_forward_pre_hooks
        ):
            return forward_call(*input, **kwargs)
        # Do not call functions when jit is used
        full_backward_hooks, non_full_backward_hooks = [], []
        if self._backward_hooks or _global_backward_hooks:
            full_backward_hooks, non_full_backward_hooks = self._get_backward_hooks()
        if _global_forward_pre_hooks or self._forward_pre_hooks:
            for hook in itertools.chain(
                _global_forward_pre_hooks.values(), self._forward_pre_hooks.values()
            ):
                result = hook(self, input)
                if result is not None:
                    if not isinstance(result, tuple):
                        result = (result,)
                    input = result

        bw_hook = None
        if full_backward_hooks:
            bw_hook = hooks.BackwardHook(self, full_backward_hooks)
            input = bw_hook.setup_input_hook(input)

        result = forward_call(*input, **kwargs)
        if _global_forward_hooks or self._forward_hooks:
            for hook in itertools.chain(
                _global_forward_hooks.values(), self._forward_hooks.values()
            ):
                hook_result = hook(self, input, result)
                if hook_result is not None:
                    result = hook_result

        if bw_hook:
            result = bw_hook.setup_output_hook(result)

        # Handle the non-full backward hooks
        if non_full_backward_hooks:
            var = result
            while not isinstance(var, torch.Tensor):
                if isinstance(var, dict):
                    var = next((v for v in var.values() if isinstance(v, torch.Tensor)))
                else:
                    var = var[0]
            grad_fn = var.grad_fn
            if grad_fn is not None:
                for hook in non_full_backward_hooks:
                    wrapper = functools.partial(hook, self)
                    functools.update_wrapper(wrapper, hook)
                    grad_fn.register_hook(wrapper)
                self._maybe_warn_non_full_backward_hook(input, result, grad_fn)

        return result

    __call__ = _call_impl
